/*
Next Version:

[ ] Output: Hex, Decimal, or binary (answer.toString(16) answer.toString(2))
[ ] Hack: input as binary '0b11011'. convert to 'bin("11011")'     "0b[01]"

Future Ideas:
[ ] "Clear Memory" (with an undo option)
[ ] Save/restore input history.
[ ] How to handle prop objects? Example: "p = {x:0, y:1}"
[ ] help in a sepearate html file?
[ ] Input: degrees or radians
[ ] Option: Do/Don't save memory
[ ] Regex to dissallow access of "window" object, other globals.

List of common functions:
dist = function (x1, y1, x2, y2) { return sqrt(pow(x2-x1, 2)+pow(y2-y1, 2)); }
sign = function (x) { return x < 0 ? -1 : x > 0 ? 1 : 0; }
*/

function Sandbox() {
	
	// Do calculations in a sandbox object that doesn't have any local variables.
	// Note: Expressions can still be malicious by accessing the "window" object.
	this.evaluate = function(_expression, answer) {
		with (Math) {
			var pi = PI;
			var e = E;
			return eval(_expression);
		}
	}
}

function Calc() {
	
	// History of inputs
	var history = new Array();
	var historyIndex = -1;
	var maxHistory = 100;
	
	var lastAnswer = 0;
	var lastError = false;
	
	var sandbox = new Sandbox();
	
	var _self = this;
	
	// List of known variables at startup. Other variables will be considered "user variables"
	var knownMembers = new Array();
	for (var i in window) {
		knownMembers.push(i);
	}
	knownMembers.push("CALC_UI");
	loadUserVars();
	
	function loadUserVars() {
		if (window.widget) {
			var memory = widget.preferenceForKey("memory");
			if (memory != undefined && memory.length > 0) {
				try {
					sandbox.evaluate(memory, 0);
				}
				catch (ex) {
					// Do something?
				}
			}
		}
	}
	
	this.clearHistory = function() {
		history = new Array();
		historyIndex = -1;
	}
	
	this.getLastAnswer = function() {
		return lastAnswer;
	}
	
	this.getUserVars = function() {
		var userVars = new Array();
		
		for (var i in window) {
			
			// Don't include prototypes
			if (i.indexOf("[[") == 0) {
				continue;
			}
			
			var isKnown = false;
			for (var j in knownMembers) {
				if (i === knownMembers[j]) {
					isKnown = true;
					break;
				}
			}
			if (!isKnown) {
				userVars.push(i);
			}
		}
		
		return userVars;
	}
	
	this.saveUserVars = function() {
		var userVars = _self.getUserVars();
		var memory = "";
		for (var i in userVars) {
			var name = userVars[i];
			if (window[name] != undefined) {
				var value = window[name].toString();
				value = value.replace(/\n/g, ' ');
				memory += name + "=" + value + ";";
			}
		}
		if (window.widget) {
			widget.setPreferenceForKey(memory, "memory");
		}
	}
	
	this.clearUserVars = function() {
		lastAnswer = 0;
		lastError = false;
		
		var userVars = _self.getUserVars();
		for (var i in userVars) {
			var name = userVars[i];
			try {
				eval("delete " + name);
			}
			catch (ex) {
				// Ignore
			}
		}
		if (window.widget) {
			widget.setPreferenceForKey(null, "memory");
		}
	}
	
	this.calc = function(expression, addToHistory) {
		
		// Add to history
		if (addToHistory) {
			history.push(expression);
			if (history.length > maxHistory) {
				history = history.slice(history.length - maxHistory); 
			}
		}
		historyIndex = history.length;
			
		// Evaluate the expression
		var answer;
		lastError = false;
		try {
			answer = sandbox.evaluate(expression, lastAnswer);
			if (answer == undefined) {
				lastError = true;
			}
			else if (typeof answer == "number") {
				lastAnswer = answer;
			}
			else if (typeof answer == "string") {
				answer = "\"" + answer + "\"";
			}
			else if (answer instanceof Array) {
				answer = "[" + answer + "]";
			}
			else if (typeof answer == "function") {
				answer = "Function defined";
			}
			else if (typeof answer != "boolean") {
				answer = "SyntaxError";
				lastError = true;
			}
		}
		catch (ex) {
			answer = ex;
			lastError = true;
		}
		
		return answer;
	}
	
	// Returns true if the last calculation in calc() resulted in an error
	this.wasError = function() {
		return lastError;
	}
	
	this.getHistoryPrev = function() {
		if (historyIndex > 0) { 
			historyIndex = Math.min(historyIndex - 1, history.length);
			return history[historyIndex];
		}
		else {
			return null;
		}
	}
	
	this.getHistoryNext = function() {
		if (historyIndex < history.length - 1) {
			historyIndex = Math.max(0, historyIndex + 1);
			return history[historyIndex];
		}
		else {
			historyIndex = history.length;
			return "";
		}
	}
}
